/*
	
Tetris for Nintendo DS

Big kudos to Joat and Dovoto for RE, DarkFader for hardware RE, and dovoto for the snazzy collection of headers.


I've already planned use for the second screen and the touchscreen, so this release is more like a proof of concept.
So don't ask about these features, they will come :P


DesktopMan
desktopman@gmail.com

*/

// lib includes from pernlib
#include <NDS/NDS.h>
#include <keys.h>

// project includes
#include "piece.h"
#include "colors.h"
#include "status.h"
#include "highscore.h"
#include "mt.h"
#include "text.h"
#include "bg.h"

// Data headers
#include "tiles.h"
#include "font.h"

#include "tilespal.h"
#include "fontpal.h"

#include "pieces.h"

void WaitForVblank();

int players=1;
//#define REG_VCOUNT			(*(vu16*)0x4000006)

// stat contains information about player
cStatus stat[2];
// piece contains information about the current dropping piece[i]. This instance is recycled when the piece lands.
cPiece piece[2];

// Move piece every timing[level] vblank. (every 1/(60/value) seconds)
int timing[]={30,27,24,21,18,15,12,9,6,3};

// Where to offset the playing board on our background
int offsx=10;
int offsy=2;

int main()
{
	
	WAIT_CR=0xe800;
	POWER_CR=0x30F;

	VRAM_A_CR=VRAM_ENABLE|VRAM_A_MAIN_BG; // bg 0x06000000 - 0x06020000
	VRAM_B_CR=VRAM_ENABLE|VRAM_B_MAIN_BG; // bg 0x06020000 - 0x06040000
	VRAM_C_CR=VRAM_ENABLE|VRAM_C_SUB_BG; // bg 0x06200000
	VRAM_D_CR=VRAM_ENABLE|VRAM_D_SUB_SPRITE; // sprites 0x06600000 (unused as of yet)

	// Is sram valid, if not, fix it.
	HighscoreInit();

	KeysInit();
	ScanKeys();

	if(KeysHeld()&KEY_A)
	{
		players=2;
	}

	piece[0].key_left=KEY_LEFT;
	piece[0].key_right=KEY_RIGHT;
	piece[0].key_rot=KEY_UP;
	piece[0].key_drop=KEY_DOWN;
	
	piece[1].key_left=KEY_Y;
	piece[1].key_right=KEY_A;
	piece[1].key_rot=KEY_X;
	piece[1].key_drop=KEY_B;

	// Text bg, 16 color, char base 1 and screen base 0
	cBg texta(0,BG_COLOR_16,1,0);
	// Copy font data to vram
	dmaCopy(font,texta.tile,(int)font_size);

	// Make a bg class instance, 256 color, char base 2 and screen base 1
	cBg boarda(1,BG_COLOR_256,2,1);
	// Copy tile data to vram
	dmaCopy(tiles,boarda.tile,(int)tiles_size);

	// Text bg, 16 color, char base 1 and screen base 0
	cBgSub textb(0,BG_COLOR_16,1,0);
	// Copy font data to vram
	dmaCopy(font,textb.tile,(int)font_size);

	// Make a bg class instance, 256 color, char base 2 and screen base 1
	cBgSub boardb(1,BG_COLOR_256,2,1);
	// Copy tile data to vram
	dmaCopy(tiles,boardb.tile,(int)tiles_size);

	cBgBase *text[2];
	text[1]=&texta;
	text[0]=&textb;

	cBgBase *board[2];
	board[1]=&boarda;
	board[0]=&boardb;

	for(int i=0;i<players;i++)
	{
		// All tiles transparent at startup, both font data and tile data has pos 0 as transparent.
		for(int pos=0;pos<1024;pos++)
		{
			board[i]->map[pos]=TRANSPARENT;
			text[i]->map[pos]=TRANSPARENT;
		}

		// Generate top border
		for(int x=0;x<10;x++)
			board[i]->map[offsx+x+32]=METAL;

		// Generate bottom border
		for(int x=0;x<10;x++)
			board[i]->map[offsx+x+22*32]=METAL;

		// Generate left border
		for(int y=1;y<23;y++)
			board[i]->map[offsx-1+y*32]=METAL;

		// Generate right border
		for(int y=1;y<23;y++)
			board[i]->map[offsx+10+y*32]=METAL;

		TextWrite(10,0,"Press Start",text[i]);
		TextWrite(5,23,"Tetris, by DesktopMan.",text[i]);
	}

	// Copy palettes to vram
	dmaCopy(tilespal,PALETTE,512);
	dmaCopy(fontpal,&PALETTE[256-16],32);
	dmaCopy(tilespal,PALETTE_SUB,512);
	dmaCopy(fontpal,&PALETTE_SUB[256-16],32);

	// Black backdrop
	PALETTE[0]=0x0;
	PALETTE_SUB[0]=0x0;

	// Wait until vblank before turning the backgrounds on.
	WaitForVblank();

	// Text mode, bgs on.
	DISPLAY_CR=MODE_0_2D|DISPLAY_BG0_ACTIVE|DISPLAY_BG1_ACTIVE;
	SUB_DISPLAY_CR=MODE_0_2D|DISPLAY_BG0_ACTIVE|DISPLAY_BG1_ACTIVE;

	ScanKeys();

	// Wait until start is not pressed, incase it is held at startup
	while((KeysHeld()&KEY_START))
	{
		WaitForVblank();
		ScanKeys();
	}

	int temp=0;

	// Wait until start is pressed again, to get a seed value
	while(!(KeysHeld()&(KEY_START|KEY_TOUCH)))
	{
		WaitForVblank();
		ScanKeys();

		temp++;
	}

	// Seed the mersenne twister with the number of vblanks waited.
	sgenrand(temp);


	for(int i=0;i<players;i++)
	{
		TextWrite(10,0,"           ",text[i]);

		// Reset the piece to a random structure
		piece[i].numnext=genrand()%7;
		piece[i].reset(genrand()%7);
		piece[i].draw();

		// Write some informal text, and the current highscore to screen.
		TextWrite(1,2,"Record",text[i]);
		TextWrite(1,3,HighscoreRead(),text[i]);

		TextWrite(23,3,"Next",text[i]);

		TextWrite(1,6,"Points",text[i]);
		TextWrite(1,9,"Lines",text[i]);
		TextWrite(1,12,"Level",text[i]);
	}

	piece[0].game=true;

	if(players==2)
		piece[1].game=true;
	else
		piece[1].game=false;

	WaitForVblank();

	// Main game loop
	while(piece[0].game||piece[1].game)
	{
		WaitForVblank();

		// Screen update
		for(int i=0;i<players;i++)
		{
			for(int y=0;y<20;y++)
			{
				for(int x=0;x<10;x++)
				{
					board[i]->map[x+offsx+(y+offsy)*32]=piece[i].board[x+y*10];
				}
			}


			// write next block to screen
			for(int y=0;y<4;y++)
			{
				for(int x=0;x<4;x++)
				{
					if(pieces[piece[i].numnext*4*4*4+x+y*4])
						board[i]->map[23+x+(y+4)*32]=piece[i].numnext+1;
					else
						board[i]->map[23+x+(y+4)*32]=TRANSPARENT;
				}
			}
		}

		stat[0].frames++;
		stat[1].frames++;

		ScanKeys();

		for(int i=0;i<players;i++)
		{
			if(piece[i].game==false)
				continue;

			// Write stats
			TextWrite(1,7,stat[i].points,text[i]);
			TextWrite(1,10,stat[i].lines,text[i]);
			TextWrite(1,13,stat[i].level+1,text[i]);

			// We are (probably) going to move the piece, so remove it first.
			piece[i].undraw();

			// Do we want to, and can we move left?
			if(KeysDown()&piece[i].key_left&&piece[i].ok(-1,0,0))
				piece[i].curx--;

			// Do we want to, and can we move right?
			if(KeysDown()&piece[i].key_right&&piece[i].ok(1,0,0))
				piece[i].curx++;

			// Do we want to, and can we rotate once?
			if(KeysDown()&piece[i].key_rot&&piece[i].cury>=0&&piece[i].ok(0,0,1))
				piece[i].SetRot(piece[i].rot+1);

			// Do we want to drop until we can't anymore, and can we drop atleast once?
			if(KeysDown()&piece[i].key_drop&&piece[i].ok(0,1,0))
			{
				while(piece[i].ok(0,1,0))
					piece[i].cury++;

				stat[i].frames=1;
				piece[i].RemoveLines();
			}

			// All moving is done, let's draw it to the internal board
			piece[i].draw();

			if(stat[i].frames%timing[stat[i].level])
				continue;

			// Code below is only executed if we have waited one wait period
			// (see timing variable in main.cpp) since last y-increase. (increase beeing downwards)

			// We are going to move the piece, so we remove it first
			piece[i].undraw();

			// Can we move the piece one square down?
			if(piece[i].ok(0,1,0))
			{
				piece[i].cury++;
				piece[i].draw();

				stat[i].drops++;
			}
			else // if we can't, new piece
			{
				piece[i].draw();
				piece[i].RemoveLines();

				stat[i].points+=((24+(3*stat[i].level))-stat[i].drops);
				stat[i].drops=0;

				// If we can't, and we can't position a new piece either, death is the only solution
				if(!piece[i].reset(genrand()%7))
				{
					piece[i].game=false;

					if(stat[i].points>HighscoreRead())
						HighscoreWrite(stat[i].points);
				}
			}
		}
	}

	// The game is now over.

	for(int i=0;i<players;i++)
	{
		// Write statistics
		TextWrite(1,7,stat[i].points,text[i]);
		TextWrite(1,10,stat[i].lines,text[i]);
		TextWrite(1,13,stat[i].level+1,text[i]);
	}

	while(!(KeysHeld()&KEY_START))
	{
		WaitForVblank();
		ScanKeys();
	}

	WaitForVblank();

	swiSoftReset();
}
